"""Addon to configparser.py. Focus on IPL handling"""

from __future__ import annotations

import datetime
import os
from copy import deepcopy
from typing import TYPE_CHECKING

from fmu.config import etc

if TYPE_CHECKING:
    from typing import Any

    from .configparserfmu import ConfigParserFMU

XFMU = etc.Interaction()
logger = XFMU.functionlogger(__name__)

# pylint: disable=protected-access
# pylint: disable=too-many-branches


class ConfigError(ValueError):
    """Exception used for config error, derived from ValueError"""


def to_ipl(
    config_parser: ConfigParserFMU,
    rootname: str = "global_variables",
    destination: str | None = None,
    template: str | None = None,
    tool: str = "rms",
) -> None:
    """Export the config as a global variables IPL and/or template
    form of the IPL.

    Args:
        rootname: Root file name without extension. An extension
            .ipl will be added for destination, and .ipl.tmpl
            for template output.
        destination (str): The output file destination (folder)
        template (str): The folder for the templated version of the
            IPL (for ERT).
        tool (str): Which section in the master to use (default is 'rms')
    """

    if not destination and not template:
        raise ConfigError("Both destination and template for IPL cannot be None.")

    if destination and not os.path.isdir(destination):
        raise ConfigError(
            'Given "destination" {} is not a directory'.format(destination)
        )
    if template and not os.path.isdir(template):
        raise ConfigError('Given "template" {} is not a directory'.format(template))

    declarations = []
    expressions_dest: list[str] = []
    expressions_tmpl: list[str] = []

    metadata = config_parser._get_sysinfo(commentmarker="//")
    declarations.extend(metadata)

    hdecl, hlist = _ipl_stringlist_format(config_parser, "horizons", tool=tool)
    if hdecl is not None:
        declarations.extend(hdecl)
        assert hlist is not None
        expressions_dest.extend(hlist)
        expressions_tmpl.extend(hlist)

    hdecl, hlist = _ipl_stringlist_format(config_parser, "zones", tool=tool)
    if hdecl is not None:
        declarations.extend(hdecl)
        assert hlist is not None
        expressions_dest.extend(hlist)
        expressions_tmpl.extend(hlist)

    hdecl, hlist = _ipl_kwlists_format(config_parser, tool=tool)
    if hdecl is not None:
        declarations.extend(hdecl)
        assert hlist is not None
        expressions_dest.extend(hlist)
        expressions_tmpl.extend(hlist)

    # freeform formats (most complex to handle)
    if destination:
        hdecl, hlist = _ipl_freeform_format(config_parser)
        if hdecl is not None:
            declarations.extend(hdecl)
            assert hlist is not None
            expressions_dest.extend(hlist)

        destfile = os.path.join(destination, rootname + ".ipl")
        with open(destfile, "w", encoding="utf-8") as stream:
            for line in declarations:
                stream.write(line)

            for line in expressions_dest:
                stream.write(line)

    if template:
        hdecl, hlist = _ipl_freeform_format(config_parser, template=True)
        if hdecl is not None:
            if not destination:
                declarations.extend(hdecl)

            assert hlist is not None
            expressions_tmpl.extend(hlist)

        tmplfile = os.path.join(template, rootname + ".ipl.tmpl")
        with open(tmplfile, "w", encoding="utf-8") as stream:
            for line in declarations:
                stream.write(line)

            for line in expressions_tmpl:
                stream.write(line)


def _ipl_stringlist_format(
    config_parser: ConfigParserFMU, subtype: str, tool: str = "rms"
) -> tuple[list[str] | None, list[str] | None]:
    """Process the rms horizons etc, and return declarations and values."""

    cfg = config_parser.config[tool].get(subtype)
    if cfg is None:
        return None, None

    decl = []
    expr = []
    for variable in cfg:
        mydecl = "String {}[]\n".format(variable)
        decl.append(mydecl)

        array = cfg[variable]
        for inum, element in enumerate(array):
            mylist = '{}[{}] = "{}"\n'.format(variable, inum + 1, element)
            expr.append(mylist)

    expr.append("\n")

    return decl, expr


def _ipl_kwlists_format(
    config_parser: ConfigParserFMU, tool: str = "rms"
) -> tuple[list[str] | None, list[str] | None]:
    """Process the rms 'kwlists', and return declarations and values.

    This format is on the form::

      rms:
        kwlists:

          FACIESNAMES:
            OFFSHORE_VI_C: [1, "Offshore mudstones, Viking Gp."]
            MUDDY_SPIC_C: [2, "Muddy spiculites"]
            BIOSTROME_REEF_C: [3, "Biostrome reef"]
            SANDY_SPIC_C: [4, "Sandy spiculites"]
            TSS_Z1_C: [5, "Transgressive sands, Draupne Fm 2"]

    It should then give::

       Int OFFSHORE_VI_C = 1
       Int MUDDY_SPIC_C = 2
       etc...

       String FACIESNAMES[]

       FACIESNAMES[OFFSHORE_VI_C] = "Offshore mudstones, Viking Gp."
       etc
    """

    cfg = config_parser.config[tool].get("kwlists")
    if cfg is None:
        return None, None

    decl = []
    expr = []
    for key, var in cfg.items():
        mydecl = "String {}[]\n".format(key)
        decl.append(mydecl)

        for subkey, (code, fullname) in var.items():
            # logger.info(subkey, code, fullname)
            mydecl = "Int {} = {}\n".format(subkey, code)
            decl.append(mydecl)

            mylist = '{}[{}] = "{}"\n'.format(key, subkey, fullname)
            expr.append(mylist)

        expr.append("\n")

    return decl, expr


def _cast_value(value: Any) -> Any:
    """Convert data type when a number is represented as a string,
    e.g. '1' or '34.33'
    """

    logger.info("Value is of type %s", type(value))
    result = value
    if isinstance(value, str):
        if "." in value:
            try:
                result = float(value)
            except ValueError:
                result = value
        elif value.lower() in ("yes", "true"):
            result = True
        elif value.lower() in ("no", "false"):
            result = False
        else:
            try:
                result = int(value)
            except ValueError:
                result = value
    else:
        result = value

    return result


def _guess_dtype(var: str, entry: dict[str, Any]) -> dict[str, Any]:
    """Guess the IPL dtype from value or values if dtype is missing.

    The entry itself will then be a scalar or a list, which need to be
    analysed. If a list, only the first value is analysed for data
    type; then it is ASSUMED it is the type...

    Returns a dict as usekey[keyword]['dtype'] and
    usekey[keyword]['value'] or usekey[keyword]['values']
    """

    values = entry[var]
    keyword = var
    logger.info("Guess dtype and value(s) for %s %s", var, values)

    usekey: dict[str, dict[str, Any]] = {}
    usekey[keyword] = {}
    usekey[keyword]["dtype"] = None
    usekey[keyword]["value"] = None  # Keep "value" if singel entry
    usekey[keyword]["values"] = None  # Keep "values", if list

    if isinstance(values, list):
        checkval = values[0]
        scheckval = str(checkval)
        if "~" in scheckval:
            val, _xtmp = scheckval.split("~")
            checkval = val.strip()
            checkval = _cast_value(checkval)

        usekey[keyword]["values"] = values
        del usekey[keyword]["value"]
    else:
        checkval = values
        scheckval = str(checkval)
        if "~" in scheckval:
            val, _xtmp = scheckval.split("~")
            checkval = val.strip()
            checkval = _cast_value(checkval)
        usekey[keyword]["value"] = values
        del usekey[keyword]["values"]

    for alt in ("int", "str", "float", "bool"):
        if alt in str(type(checkval)):
            usekey[keyword]["dtype"] = alt
            break

    if not usekey[keyword]["dtype"]:
        # dtype is still None; evaluate for date or datepair:
        if isinstance(checkval, list):
            checkval = checkval[0]
            if isinstance(checkval, datetime.date):
                usekey[keyword]["dtype"] = "datepair"
        else:
            if isinstance(checkval, datetime.date):
                usekey[keyword]["dtype"] = "date"

    # final check
    if not usekey[keyword]["dtype"]:
        raise RuntimeError("Cannot find dtype")

    logger.info("Updated key XX dtype is %s", usekey[keyword]["dtype"])
    logger.info("Updated key XX is %s", usekey)
    return usekey


def _ipl_freeform_format(
    config_parser: ConfigParserFMU, template: bool = False
) -> tuple[list[str] | None, list[str] | None]:
    """Process the RMS IPL YAML config freeform types.

    The freeform types are e.g. like this::

        rms:
          KH_MULT_MTR:
            dtype: float
            value: 1.0 ~ <KH_MULT_MTR>  # <..> be used in ERT template

          GOC:
            dtype: float
            values:
              - 2010.0
              - 2016.0

    I.e. they are defined as *UPPERCASE_LETTER* keys within
    the RMS section, in contrast to 'horizons' and 'zones'

    Args:
        template (bool): If True, then the tvalue* are returned, if present

    """

    decl = ["\n// Declare free form:\n"]
    expr = ["\n// Free form expressions:\n"]

    cfg = config_parser.config["rms"]

    # collect uppercase keys in 'rms'
    freeform_keys = []
    for key in cfg:
        if all(word[0].isupper() for word in key if word.isalpha()):
            freeform_keys.append(key)

    if not freeform_keys:
        return None, None

    for variable in freeform_keys:
        logger.info("Variable to process is %s", variable)
        expr.append("\n")

        if variable.startswith("_IPL_CODE"):
            logger.info("IPL code stub: %s \n%s", variable, cfg[variable])
            expr.append(cfg[variable])
            continue

        if variable.startswith("_IPL_DECLARE"):
            logger.info("IPL declare only: %s \n%s", variable, cfg[variable])
            decl.append(cfg[variable])
            continue

        if not isinstance(cfg[variable], dict):
            guesscfg = _guess_dtype(variable, cfg)

            usecfg = guesscfg[variable]
        else:
            usecfg = deepcopy(cfg[variable])

        mydtype = usecfg["dtype"]
        myvalue = usecfg.get("value")
        myvalues = usecfg.get("values")

        if mydtype in ("date", "datepair"):
            if myvalue is not None:
                raise RuntimeError(
                    '<{}>: Treating <date> as "value" is not '
                    'possible, rather make into list "values" '
                    "with one entry instead!".format(myvalue)
                )
            myvalues = _fix_date_format(variable, mydtype, myvalues, aslist=True)
            mydtype = "str"

        if myvalue is None and myvalues is None:
            raise ConfigError(
                "'value' or 'values' is missing for RMS variable {}".format(variable)
            )

        adecl, aexpr = _freeform_handle_entry(
            variable, myvalue, myvalues, mydtype, template
        )

        logger.info("Append %s", adecl)
        decl.append(adecl)
        expr.append(aexpr)

    decl.append("//{} {}\n\n".format("-*- END IPL DECLARATIONS -*-", "-" * 48))

    return decl, expr


def _freeform_handle_entry(
    variable: str,
    myvalue: str | bool | None,
    myvalues: list[str] | None,
    dtype: str,
    template: bool,
) -> tuple[str, str]:  # pylint: disable=too-many-statements
    """Handling of any entry as single value or list in IPL.

    Either myvalue or myvalues shall be None!

    Args:
        variable (str): Name of variable
        myvalue (str or bool): Single value case
        myvalues (): List of values
        dtype (str): Type of variable on python form: "str", "float", "int", ...

    Returns:
        (decl, expr): Declaration string and expression string
    """

    decl = ""
    expr = ""

    subtype = None
    if dtype is not None:
        subtype = dtype.capitalize()
    if "Str" in subtype:
        subtype = "String"

    logger.info(
        "Subtype for %s is %s (myvalue is %s, dtype is %s)",
        variable,
        subtype,
        myvalue,
        dtype,
    )

    # inner function
    def _fixtheentry(
        variable: str,
        myval: str | bool,
        subtype: str,
        count: int | None = None,
        template: bool = False,
    ) -> tuple[str, str]:
        logger.info("Fix freeform entry %s (subtype %s)", variable, subtype)
        tmpvalue = str(myval)
        if "~" in tmpvalue:
            val, var = tmpvalue.split("~")
            val = val.strip()
            var = var.strip()
        else:
            val = tmpvalue.strip()
            var = None

        if subtype == "Bool":
            if val in ("True", "yes", "YES", "Yes", "true", "TRUE"):
                val = "TRUE"
            if val in ("False", "no", "NO", "No", "false", "FALSE"):
                val = "FALSE"

        if subtype == "Float":
            logger.info("Input float value is %s (%s)", val, variable)
            if "e" in str(val).lower():
                val = "{0:E}".format(float(val))

            logger.info("Updated float value is %s (%s)", val, variable)

        if subtype == "String":
            val = '"{}"'.format(val)
            if var:
                var = '"{}"'.format(var)

        myvalue = val
        if var:
            myvalue = val + "  // " + var

        if var and template:
            myvalue = var + "  // " + val

        counter = ""
        if count:
            counter = "[" + str(count) + "]"
        expr = variable + counter + " = " + myvalue + "\n"

        decltype = ""
        if count:
            decltype = "[]"  # e.g. Bool SOMEBOOL[]

        decl = subtype + " " + variable + decltype + "\n"

        return decl, expr

    # single entry
    if myvalue is not None:
        decl, expr = _fixtheentry(variable, myvalue, subtype, template=template)

    # list entry
    elif myvalues is not None:
        expr = ""
        for icount, myval in enumerate(myvalues):
            decl, subexpr = _fixtheentry(
                variable, myval, subtype, count=icount + 1, template=template
            )
            expr += subexpr

    return decl, expr


def _fix_date_format(
    var: str, dtype: str, value: str | None, aslist: bool = False
) -> str | list[str] | None:
    """Make dateformat to acceptable RMS IPL format."""

    logger.info("Fix dates for %s", var)
    if value is None:
        returnv = None

    logger.info("Fix dates... dtype is %s", dtype)
    if dtype not in ("date", "datepair"):
        logger.info("Fix dates... dtype is %s RETURN", dtype)
        return value

    values: list[tuple[datetime.datetime, datetime.datetime]] | str | None = None
    if aslist:
        logger.debug("Dates is a list")
        values = value
        value = ""

    if dtype == "date":
        logger.info("Process date ...")
        if values:
            mynewvalues: list[str] = []
            logger.info("Process date as values")
            for val in values:
                if isinstance(val, (datetime.datetime, datetime.date)):
                    val = str(val)
                    val = val.replace("-", "")
                    mynewvalues.append(val)
            returnv = mynewvalues

    if dtype == "datepair" and values:
        mynewvalues = []
        for val in values:
            date1, date2 = val  # type: ignore
            # -- might be iterable, list, tuple
            if isinstance(date1, (datetime.datetime, datetime.date)):
                date1 = str(date1)
                date1 = date1.replace("-", "")
            if isinstance(date2, (datetime.datetime, datetime.date)):
                date2 = str(date2)
                date2 = date2.replace("-", "")
            mynewvalues.append(date1 + "_" + date2)

        returnv = mynewvalues

    return returnv



"""Testing fmu-config."""

import json
import os.path
from os.path import join

import pytest
import yaml

import fmu.config as config
from fmu.config import utilities as ut

# import fmu.config.fmuconfigrunner as fmurun

JFILE1 = "tests/data/yml/vinstre/global_master_config.yml"
TFILE2 = "tests/data/yml/ogre/global_master_config.yml"
TFILE3 = "tests/data/yml/ogre/global_master_config_with_dupl.yml"
TFILE4 = "tests/data/yml/ogre/global_master_config_with_merge.yml"
RFILE1 = "tests/data/yml/reek/global_variables.yml"

# result may be compared with data stored here
TCMP = "tests/data/yml/test_compare"

fmux = config.etc.Interaction()
logger = fmux.basiclogger(__name__)

# always this statement
if not fmux.testsetup():
    raise SystemExit

TMPD = fmux.tmpdir


def test_vinstre():
    """Test table output behaviour"""

    cfg = config.ConfigParserFMU()
    cfx = config.ConfigParserFMU()

    assert isinstance(cfg, config.ConfigParserFMU)

    cfg.parse(JFILE1)
    cfx.parse(JFILE1, smart_braces=False)

    cfg.to_table(
        rootname="VI",
        destination=fmux.tmpdir,
        template=fmux.tmpdir,
        entry="global.FWL",
        sep="      ",
    )

    cfg.to_ipl(
        rootname="vinstre_global_variables",
        destination=fmux.tmpdir,
        template=fmux.tmpdir,
        tool="rms",
    )

    cfg.to_yaml(
        rootname="vinstre_global_variables_rms",
        destination=fmux.tmpdir,
        template=fmux.tmpdir,
        tool="rms",
    )

    status1 = ut.compare_yaml_files(
        join(TMPD, "vinstre_global_variables_rms.yml"),
        join(TCMP, "vinstre_global_variables_rms.yml"),
    )
    status2 = ut.compare_yaml_files(
        join(TMPD, "vinstre_global_variables_rms.yml.tmpl"),
        join(TCMP, "vinstre_global_variables_rms.yml.tmpl"),
    )
    assert status1 is True
    assert status2 is True

    cfx.to_yaml(
        rootname="vinstre_global_variables_rms_nobraces",
        destination=fmux.tmpdir,
        template=fmux.tmpdir,
        tool="rms",
    )

    assert cfx.config["rms"]["FWL3"][1] == "1236.0 ~ <>"
    assert cfg.config["rms"]["FWL3"][1] == "1236.0 ~ <FWL3_1>"


def test_basic_ogre():
    """Test basic behaviour"""

    cfg = config.ConfigParserFMU()

    assert isinstance(cfg, config.ConfigParserFMU)

    cfg.parse(TFILE2)

    # cfg.show()

    assert len(cfg.config["rms"]["horizons"]) == 6


def test_to_yaml_ogre():
    """Test the output for the YAML files, both templated and normal for rms"""

    cfg = config.ConfigParserFMU()

    assert isinstance(cfg, config.ConfigParserFMU)

    cfg.parse(TFILE2)
    rootn = "ogre_yaml"

    cfg.to_yaml(
        rootname=rootn, destination=fmux.tmpdir, template=fmux.tmpdir, tool="rms"
    )

    # now read the files again to assert tests
    with open(
        os.path.join(fmux.tmpdir, rootn + ".yml"), "r", encoding="utf-8"
    ) as stream:
        cfg_yml = yaml.safe_load(stream)

    with open(
        os.path.join(fmux.tmpdir, rootn + ".yml.tmpl"), "r", encoding="utf-8"
    ) as stream:
        cfg_tmpl = yaml.safe_load(stream)

    assert cfg_yml["KH_MULT_CSAND"] == 1.0
    assert cfg_tmpl["KH_MULT_CSAND"] == "<KH_MULT_CSAND>"

    status1 = ut.compare_yaml_files(
        join(TMPD, rootn + ".yml"),
        join(TCMP, rootn + ".yml"),
    )
    status2 = ut.compare_yaml_files(
        join(TMPD, rootn + ".yml.tmpl"),
        join(TCMP, rootn + ".yml.tmpl"),
    )
    assert status1 is True
    assert status2 is True

    # IPL version
    rootn = "ogre_ipl"
    cfg.to_ipl(
        rootname=rootn, destination=fmux.tmpdir, template=fmux.tmpdir, tool="rms"
    )

    status1 = ut.compare_text_files(
        join(TMPD, rootn + ".ipl"),
        join(TCMP, rootn + ".ipl"),
    )
    status2 = ut.compare_text_files(
        join(TMPD, rootn + ".ipl.tmpl"),
        join(TCMP, rootn + ".ipl.tmpl"),
    )
    assert status1 is True
    assert status2 is True


def test_to_yaml_ogre_selfread():
    """Test the output for the YAML files, and convert OUTPUT yaml to ipl"""

    cfg = config.ConfigParserFMU()

    cfg.parse(TFILE2)
    rootn = "ogre_yaml"

    cfg.to_yaml(rootname=rootn, destination=fmux.tmpdir, template=fmux.tmpdir)

    newinput = os.path.join(fmux.tmpdir, rootn + ".yml")
    newrootn = "ogre_yaml_selfread"

    cfx = config.ConfigParserFMU()
    cfx.parse(newinput)

    cfx.to_ipl(
        rootname=newrootn, destination=fmux.tmpdir, template=fmux.tmpdir, tool="rms"
    )


def test_ogre_to_yaml_merge():
    """Test the output for the YAML files, and convert OUTPUT yaml to ipl"""

    cfg = config.ConfigParserFMU()

    cfg.parse(TFILE4)
    rootn = "ogre_yaml"

    cfg.to_yaml(rootname=rootn, destination=fmux.tmpdir, template=fmux.tmpdir)

    newinput = os.path.join(fmux.tmpdir, rootn + ".yml")
    newrootn = "ogre_yaml_selfread"

    cfx = config.ConfigParserFMU()
    cfx.parse(newinput)

    cfx.to_ipl(
        rootname=newrootn, destination=fmux.tmpdir, template=fmux.tmpdir, tool="rms"
    )


def test_yaml_has_duplicates_ogre():
    """The YAML file has duplicates; should raise error"""

    cfg = config.ConfigParserFMU()

    with pytest.raises(SystemExit):
        cfg.parse(TFILE3)


def test_to_json_ogre():
    """Test the output for the JSON files, both templated and normal for
    rms section.
    """

    cfg = config.ConfigParserFMU()

    assert isinstance(cfg, config.ConfigParserFMU)

    cfg.parse(TFILE2)
    rootn = "ogre_json"

    cfg.to_json(
        rootname=rootn, destination=fmux.tmpdir, template=fmux.tmpdir, tool="rms"
    )

    with open(
        os.path.join(fmux.tmpdir, rootn + ".json"), "r", encoding="utf-8"
    ) as myfile:
        cfg_json = json.load(myfile)

    assert cfg_json["KH_MULT_CSAND"] == str(1.0)


def test_ipl_ogre():
    """Test basic behaviour"""

    cfg = config.ConfigParserFMU()

    assert isinstance(cfg, config.ConfigParserFMU)

    cfg.parse(TFILE2)

    # cfg.show()
    # cfg.show(style='json')

    # export the config as a global variables IPL
    logger.info("Test dir is %s", fmux.tmpdir)
    cfg.to_ipl(
        destination=os.path.join(fmux.tmpdir), template=os.path.join(fmux.tmpdir)
    )


def test_special_chars():
    """Test that special characters are carried through to produced files.

    When using non-ascii in source global_variables or include files, they shall not be
    converted to hex.
    """

    cfg = config.ConfigParserFMU()
    cfg.parse(JFILE1)
    cfg.to_yaml(
        rootname="special_chars_global_variables",
        destination=fmux.tmpdir,
        template=fmux.tmpdir,
    )
    cfg.to_json(
        rootname="special_chars_global_variables",
        destination=fmux.tmpdir,
    )

    # test assumption
    assert cfg.config["_special_chars"]["norwegian_letters"] == "æøåÆØÅ"

    # YAML
    outpath_yml = join(TMPD, "special_chars_global_variables.yml")
    assert os.path.isfile(outpath_yml)

    # open with yaml reader (which handles ÆØÅ)
    with open(outpath_yml, "r", encoding="utf-8") as stream:
        outcontent = yaml.safe_load(stream)
    assert outcontent["_special_chars"]["norwegian_letters"] == "æøåÆØÅ"

    # confirm correct encoding in raw txt
    with open(outpath_yml, "r", encoding="utf-8") as stream:
        outcontent_yaml_raw = stream.read()
    assert "_special_chars:" in outcontent_yaml_raw  # test assumption
    assert "æøåÆØÅ" in outcontent_yaml_raw

    # JSON
    outpath_json = join(TMPD, "special_chars_global_variables.json")
    assert os.path.isfile(outpath_json)

    # confirm correct encoding in raw txt
    with open(outpath_json, "r", encoding="utf-8") as stream:
        outcontent_json_raw = stream.read()
    assert "æøåÆØÅ" in outcontent_json_raw


# def test_basic_reek():
#     """Test basic behaviour, Reek setup"""

#     cfg = config.ConfigParserFMU()

#     assert isinstance(cfg, config.ConfigParserFMU)

#     cfg.parse(RFILE1)

#     cfg.show()

#     # will write the eclipse files
#     cfg.to_eclipse()
#     cfg.to_ipl()
#     cfg.to_yaml('myyaml', destination='TMP')
#     cfg.to_json('myjson', destination='TMP')

#     # assert len(cfg.config['horizons']) == 6

#     # # export the config as a global variables IPL
#     # cfg.to_ipl('myfile')


# def test_command_make_ipls():
#     """Make IPL both global_variable.ipl and global_variables.tmpl, Reek."""
#     fmurun.main(['--input', RFILE1, '--mode', 'ipl'])  # noqa


# def test_command_make_yamls():
#     """Make IPL both global_variables_rms.yml and global_variables_tmpl.yml,
#     Reek.
#     """

#     fmurun.main(['--input', RFILE1, '--mode', 'yaml', --tool, 'rms'])  # noqa



"""Module for parsing global configuration for FMU.

This module will normally be ran from the `fmuconfig` script,
which is the front-end script for the user.
"""

from __future__ import annotations

import datetime
import errno
import getpass
import json
import os
import re
import socket
import sys
from collections import Counter
from copy import deepcopy
from os.path import join as ojoin
from typing import TYPE_CHECKING

try:
    from fmu.config.version import __version__
except ImportError:
    __version__ = "0.0.0"

import yaml

from fmu.config import _configparserfmu_ipl, etc
from fmu.config._loader import ConstructorError, FmuLoader

if TYPE_CHECKING:
    from typing import Any, Literal

xfmu = etc.Interaction()
logger = xfmu.functionlogger(__name__)


class ConfigParserFMU:
    """Class for parsing global config files for FMU."""

    def __init__(self) -> None:
        self._config: dict[str, Any] = {}
        self._yamlfile: str | None = None
        self._runsilent = True
        logger.debug("Ran __init__")

    @property
    def config(self) -> dict[str, Any]:
        """Get the current config as a Python dictionary (read only)."""
        return self._config

    @property
    def yamlfile(self) -> str | None:
        """The name of the input master YAML formatted file (read only)."""
        return self._yamlfile

    def parse(self, yfile: str, smart_braces: bool = True) -> None:
        """Parsing the YAML file (reading it)."""

        with open(yfile, encoding="utf-8") as stream:
            try:
                self._config = yaml.load(stream, Loader=FmuLoader)
            except ConstructorError as errmsg:
                xfmu.error(str(errmsg))
                raise SystemExit from errmsg

        self._yamlfile = yfile

        if smart_braces:
            res = self._fill_empty_braces(deepcopy(self._config), "X")
            assert isinstance(res, dict)
            self._config = res

        self._cleanify_doubleunderscores()
        self._validate_unique_tmplkeys()

    def show(self, style: Literal["yaml", "yml", "json", "jason"] = "yaml") -> None:
        """Show (print) the current configuration to STDOUT.

        Args:
            style: Choose between 'yaml' (default), or 'json'"""

        xfmu.echo("Output of configuration:")
        if style in ("yaml", "yml"):
            yaml.dump(self.config, stream=sys.stdout)
        elif style in ("json", "jason"):
            stream = json.dumps(self.config, indent=4, default=str)
            print(stream)

    def to_table(
        self,
        rootname: str = "myconfig",
        destination: str | None = None,
        template: str | None = None,
        entry: str | None = None,
        createfolders: bool = False,
        sep: str = ",",
    ) -> None:
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-branches

        """Export a particular entry in config as text table files;
        one with true values and one with templated variables.

        Args:
            rootname: Root file name without extension. An extension
                .txt will be added for destination, and .txt.tmpl
                for template output.
            destination: The directory path for the destination
                file. If None, then no output will be given
            template: The directory path for the templated
                file. If None, then no templated output will be given.
            entry (str): Using one of the specified key/entry sections in the
                master config that holds a table, e.g. 'global.FWL'.
            createfolders (bool): If True then folders will be created if they
                do not exist (default is False).
            sep (str): Table separator, e.g. ' ', default is ','

        Raises:
            ValueError: If both destination and template output is None,
                or folder does not exist in advance, if createfolder=False,
                or entry is not spesified.

        Example:

            >>> config.to_table('fwl', destination='../',
                                entry='global.FWL')
        """

        if not destination and not template:
            raise ValueError(
                "Both destination and template are None."
                "At least one of them has to be set!."
            )

        if entry is None:
            raise ValueError("The entry is None; need a value, " 'e.g. "global.FWL"')

        if createfolders:
            self._force_create_folders([destination, template])
        else:
            self._check_folders([destination, template])

        keys = entry.split(".")

        if len(keys) == 1:
            cfg = self.config[keys[0]]
        elif len(keys) == 2:
            cfg = self.config[keys[0]][keys[1]]
        elif len(keys) == 3:
            cfg = self.config[keys[0]][keys[1]][keys[2]]
        elif len(keys) == 4:
            cfg = self.config[keys[0]][keys[1]][keys[2]][keys[3]]
        else:
            raise ValueError("Entry with more that 4 sublevels, not supported")

        if destination:
            with open(
                ojoin(destination, rootname + ".txt"), "w", encoding="utf-8"
            ) as dest:
                for row in cfg:
                    for col in row:
                        stream = self._get_required_form(str(col), template=False)
                        print(str(stream) + sep, file=dest, end="")
                    print("", file=dest)
        if template:
            with open(
                ojoin(template, rootname + ".txt.tmpl"), "w", encoding="utf-8"
            ) as tmpl:
                for row in cfg:
                    for col in row:
                        stream = self._get_required_form(str(col), template=True)
                        print(str(stream) + sep, file=tmpl, end="")
                    print("", file=tmpl)

    def to_yaml(
        self,
        rootname: str = "myconfig",
        destination: str | None = None,
        template: str | None = None,
        tool: str | None = None,
        createfolders: bool = False,
    ) -> None:
        # pylint: disable=too-many-arguments

        """Export the config as YAML files; one with true values and
        one with templated variables.

        Args:
            rootname: Root file name without extension. An extension
                .yml will be added for destination, and .yml.tmpl
                for template output.
            destination: The directory path for the destination
                file. If None, then no output will be given
            template: The directory path for the templated
                file. If None, then no templated output will be given.
            tool (str): Using one of the specified tool sections in the
                master config, e.g. 'rms'. Default is None which means all.
            createfolders (bool): If True then folders will be created if they
                do not exist (default is False).

        Raises:
            ValueError: If both destination and template output is None,
                or folder does not exist in advance, if createfolder=False.

        Example:

            >>> config.to_yaml('global_variables', destination='../')
        """
        logger.info("To YAML")
        if not destination and not template:
            raise ValueError(
                "Both destination and template are None."
                "At least one of them has to be set!."
            )

        if createfolders:
            self._force_create_folders([destination, template])
        else:
            self._check_folders([destination, template])

        # remove dtype, value(s) from RMS/IPL freeform entries
        newcfg = self._strip_rmsdtype()

        if tool is not None:
            mystream = yaml.dump(newcfg[tool], allow_unicode=True)
        else:
            mystream = yaml.dump(newcfg, allow_unicode=True)

        mystream = "".join(self._get_sysinfo()) + mystream

        mystream = re.sub(r"\s+~", "~", mystream)
        mystream = re.sub(r"~\s+", "~", mystream)

        cfg1 = self._get_dest_form(mystream)
        cfg2 = self._get_tmpl_form(mystream)
        assert isinstance(cfg1, str)  # Above can also return list
        assert isinstance(cfg2, str)

        if destination:
            out = os.path.join(destination, rootname + ".yml")
            with open(out, "w", encoding="utf-8") as stream:
                stream.write(cfg1)

        if template:
            out = os.path.join(template, rootname + ".yml.tmpl")
            with open(out, "w", encoding="utf-8") as stream:
                stream.write(cfg2)

    def to_json(
        self,
        rootname: str,
        destination: str | None = None,
        template: str | None = None,
        createfolders: bool = False,
        tool: str | None = None,
    ) -> None:
        """Export the config as JSON files; one with true values and
        one with templated variables.

        Args:
            rootname: Root file name without extension. An extension
                .json will be added for destination, and .json.tmpl
                for template output.
            destination: The directory path for the destination
                file. If None, then no output will be given
            template: The directory path for the templated
                file. If None, then no output will be given
            tool (str): Using one of the specified tool sections in the
                master config, e.g. 'rms'. Default is None which means all.
            createfolders: If True then folders will be created if they
                do not exist.

        Raises:
            ValueError: If both destination and template output is None,
                or folder does not exist in advance, if createfolder=False.

        Example:

            >>> config.to_json('global_variables', destination='../')
        """

        if not destination and not template:
            raise ValueError(
                "Both destionation and template are None."
                "At least one of them has to be set!."
            )

        if createfolders:
            self._force_create_folders([destination, template])
        else:
            self._check_folders([destination, template])

        # remove dtype, value(s) from RMS/IPL freeform entries
        newcfg = self._strip_rmsdtype()

        mycfg = newcfg[tool] if tool is not None else newcfg

        mystream = json.dumps(mycfg, indent=4, default=str, ensure_ascii=False)

        mystream = re.sub(r"\s+~", "~", mystream)
        mystream = re.sub(r"~\s+", "~", mystream)

        if destination:
            cfg1 = self._get_dest_form(mystream)
            assert isinstance(cfg1, str)  # Above can also return list
            out = os.path.join(destination, rootname + ".json")
            with open(out, "w", encoding="utf-8") as stream:
                stream.write(cfg1)

        if template:
            cfg2 = self._get_tmpl_form(mystream)
            assert isinstance(cfg2, str)  # Above can also return list
            out = os.path.join(template, rootname + ".json.tmpl")
            with open(out, "w", encoding="utf-8") as stream:
                stream.write(cfg2)

    def to_ipl(
        self,
        rootname: str = "global_variables",
        destination: str | None = None,
        template: str | None = None,
        createfolders: bool = False,
        tool: str = "rms",
    ) -> None:
        """Export the config as a global variables IPL and/or template.

        Args:
            rootname (str): Root file name without extension. An extension
                `.ipl` will be added for destination, and `.ipl.tmpl`
                for template output.
            destination (str): The output file destination (folder).
            template (str): The folder for the templated version of the
                IPL (for ERT).
            createfolders: If True then folders will be created if they
                do not exist.
            tool (str): Which section in the master to use (default is 'rms').

        """

        if createfolders:
            self._force_create_folders([destination, template])
        else:
            self._check_folders([destination, template])

        # keep most code in separate file ( a bit lengthy)
        _configparserfmu_ipl.to_ipl(
            self,
            rootname=rootname,
            destination=destination,
            template=template,
            tool=tool,
        )

    def to_eclipse(self) -> None:
        """Export the config templates and actuals under `eclipse`"""

        cfg = self.config

        for deck in cfg["eclipse"]:
            logger.info("Deck is %s", deck)
            edeck = cfg["eclipse"][deck]

            content = edeck["content"]
            content_dest = self._get_dest_form(content)
            content_tmpl = self._get_tmpl_form(content)
            assert isinstance(content_dest, str)  # Above can return list
            assert isinstance(content_tmpl, str)

            with open(edeck["destfile"], "w", encoding="utf-8") as dest:
                dest.write(content_dest)

            with open(edeck["tmplfile"], "w", encoding="utf-8") as tmpl:
                tmpl.write(content_tmpl)

    # =========================================================================
    # Private methods
    # =========================================================================

    def _cleanify_doubleunderscores(self) -> None:
        """Remove keys with double underscore in level 2, and
        move data up one level.

        This is done in order to allow anonymous include, e.g.::

           rms:
              __tmp1: !include facies.yml

        The input in facies.yaml will then be relocated to the key 'rms',
        up one level.

        .. versionchanged:: 1.0.1 secure same order of __xxx keys

        """

        # pylint: disable=too-many-nested-blocks
        newcfg = deepcopy(self._config)

        for key, val in self._config.items():
            if isinstance(val, dict):
                subkeyorder = []
                _tmps = {}

                for subkey, subval in val.items():
                    if subkey.startswith("__"):
                        if isinstance(subval, dict):
                            for subsubkey, subsubval in subval.items():
                                _tmps[subsubkey] = deepcopy(subsubval)
                                subkeyorder.append(subsubkey)
                        del newcfg[key][subkey]
                    else:
                        subkeyorder.append(subkey)
                        _tmps[subkey] = subval

                ordered = {}
                for keyw in subkeyorder:
                    ordered[keyw] = _tmps[keyw]
                newcfg[key] = ordered

        self._config = newcfg

    def _validate_unique_tmplkeys(self) -> None:
        """Collect all <...> and check that they are unique and uppercase.

        Note that duplicate <xxx> may be OK, and a print should only be issued
        if required, as information.

        """

        mystream = yaml.dump(self._config)
        tlist = []
        tmpl = re.findall(r"<\w+>", mystream)
        for item in tmpl:
            tlist.append(item)

        for item in tlist:
            wasitem = item
            item = item.rstrip(">")
            item = item.lstrip("<")
            if any(char.islower() for char in item):
                xfmu.error(
                    "Your template key contains lowercase " "letter: {}".format(wasitem)
                )

        if len(tlist) != len(set(tlist)) and not self._runsilent:
            xfmu.echo("Note, there are duplicates in <...> keywords")
            counter = Counter(tlist)
            for item, cnt in counter.items():
                if cnt > 1:
                    xfmu.echo("{0:30s} occurs {1:3d} times".format(item, cnt))

    def _fill_empty_braces(
        self, stream: str | list | dict, key: str
    ) -> str | list | dict[str, Any]:
        """If an empty variable is given, this shall be replaced with
        key name.

        For example::

              FLW1: 2000~<>

        shall be::

              FLW1: 2000~<FWL1>

        This function uses recursion

        Args:
            stream: a dict or string or...
            key: To use as <KEY>

        """
        if isinstance(stream, str):
            if key in ("value", "values") and "<>" in stream:
                xfmu.warn(
                    'Empty template "<>" is not supported in "value" or '
                    '"values" fields: {}'.format(stream)
                )
            else:
                return stream.replace("<>", "<" + str(key) + ">")
        elif isinstance(stream, list):
            return [
                self._fill_empty_braces(item, str(key) + "_" + str(num))
                for num, item in enumerate(stream)
            ]
        elif isinstance(stream, dict):
            return {
                xkey: self._fill_empty_braces(item, xkey)
                for xkey, item in stream.items()
            }
        return stream

    @staticmethod
    def _get_sysinfo(commentmarker: str = "#") -> list[str]:
        """Return a text string that serves as info for the outpyt styles
        that support comments."""

        host = socket.gethostname()
        user = getpass.getuser()
        now = str(datetime.datetime.now())
        ver = __version__
        cmt = commentmarker

        return [
            "{} Autogenerated from global configuration.\n".format(cmt),
            "{} DO NOT EDIT THIS FILE MANUALLY!\n".format(cmt),
            "{} Machine {} by user {}, at {}, using fmu.config ver. {}\n".format(
                cmt, host, user, now, ver
            ),
        ]

    @staticmethod
    def _force_create_folders(folderlist: list[str | None]) -> None:
        for folder in folderlist:
            if folder is None:
                continue
            try:
                os.makedirs(folder)
            except OSError as errmsg:
                if errmsg.errno != errno.EEXIST:
                    raise

    @staticmethod
    def _check_folders(folderlist: list[str | None]) -> None:
        for folder in folderlist:
            if folder is None:
                continue

            if not os.path.exists(folder):
                raise ValueError(
                    "Folder {} does not exist. It must either "
                    "exist in advance, or the createfolders key"
                    "must be True.".format(folder)
                )

    def _strip_rmsdtype(self) -> dict[str, Any]:
        """Returns a copy of the _config dictionary so that the (e.g.)
        FREEFORM['dtype'] and FREEFORM['value'] = x or FREEFORM['values'] = x
        becomes simplified to FREEFORM = x
        """

        newcfg = deepcopy(self._config)

        if "rms" in self._config:
            cfgrms = self._config["rms"]
        else:
            return newcfg

        for key, val in cfgrms.items():
            logger.debug(key, val)
            if isinstance(val, dict):
                logger.debug(val.keys())
                if "dtype" and "value" in val:
                    newcfg["rms"][key] = deepcopy(val["value"])
                elif "dtype" and "values" in val:
                    newcfg["rms"][key] = deepcopy(val["values"])
                elif "dtype" in val:
                    raise RuntimeError(
                        'Wrong input YAML?. It seems that "{}" has '
                        '"dtype" but no "value" or "values" ({})'.format(
                            key, val.keys()
                        )
                    )

        return newcfg

    @staticmethod
    def _get_required_form(
        stream: str | list, template: bool = False, ipl: bool = False
    ) -> list | str | None:
        """Given a variable on form 1.0 ~ <SOME>, return the required form.

        For single values, it should be flexible how it looks like, e.g.::

            19.0~<SOME>
            19.0 ~ <SOME>

        If ipl is True and template is false, it should return::

            19.0  // <SOME>

        If ipl is True and template is True, it should return::

            <SOME>  // 19.0

        Args:
            stream (:obj:`str` or :obj:`list` of :obj:`str`): Input string
                or list of strings to break up
            template (bool): If template mode
            ipl (bool): If IPL mode
        """

        if "~" not in stream:
            return stream

        result = None

        if isinstance(stream, list):
            pass
        elif isinstance(stream, str):
            if "~" in stream:
                value, tvalue = stream.split("~")
                value = value.strip()
                tvalue = tvalue.strip()

                if ipl:
                    if template:
                        result = tvalue + "  // " + value
                    else:
                        result = value + "  // " + tvalue
                else:
                    result = tvalue if template else value

        else:
            raise ValueError("Input for templateconversion neither string " "or list")

        return result

    @staticmethod
    def _get_tmpl_form(stream: list | str) -> list | str:
        """Get template form (<...> if present, not numbers)."""

        pattern = "\\-*[\\-a-zA-Z0-9._/]+~"

        if isinstance(stream, list):
            logger.info("STREAM is a list object")
            result = []
            for item in stream:
                moditem = re.sub(pattern, "", item)
                moditem = re.sub('"', "", moditem)
                moditem = " " + moditem.strip()
                result.append(moditem)
            return result

        if isinstance(stream, str):
            logger.info("STREAM is a str object - get tmpl form")
            res = re.sub(pattern, "", stream)
            res = re.sub('"', "", res)
            return res.strip() + "\n"

        raise ValueError("Input for templateconversion neither string or list")

    @staticmethod
    def _get_dest_form(stream: list | str) -> list | str:
        """Get destination form (numbers, not <...>)"""

        logger.info("TRY DEST %s", stream)
        pattern = "~.*<.+?>"

        if isinstance(stream, list):
            logger.info("STREAM is a list object")
            result = []
            for item in stream:
                moditem = re.sub(pattern, "", item)
                moditem = " " + moditem.strip()
                result.append(moditem)
            return result

        if isinstance(stream, str):
            logger.info("STREAM is a str object - get dest form")
            res = re.sub(pattern, "", stream)
            return res.strip() + "\n"

        raise ValueError("Input for templateconversion neither string " "or list")



"""Loading nested configs for FMU"""

from __future__ import annotations

import os.path
from typing import TYPE_CHECKING

import yaml
from yaml.constructor import ConstructorError
from yaml.nodes import MappingNode

if TYPE_CHECKING:
    from types_pyyaml import _Node, _ReadStream


class FmuLoader(yaml.Loader):
    """Class for making it possible to use nested YAML files.

    Code is borrowed from David Hall:
    https://davidchall.github.io/yaml-includes.html
    """

    # pylint: disable=too-many-ancestors

    def __init__(self, stream: _ReadStream) -> None:
        self._root = os.path.split(stream.name)[0]
        super(FmuLoader, self).__init__(stream)

        FmuLoader.add_constructor(
            yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, FmuLoader.construct_mapping
        )

        FmuLoader.add_constructor("!include", FmuLoader.include)
        FmuLoader.add_constructor("!import", FmuLoader.include)

    def include(self, node: _Node) -> list | dict | None:
        """Include method"""

        result: list | dict | None = None
        if isinstance(node, yaml.ScalarNode):
            scalar = self.construct_scalar(node)
            assert isinstance(scalar, str)
            result = self.extract_file(scalar)

        elif isinstance(node, yaml.SequenceNode):
            result = []
            for filename in self.construct_sequence(node):
                result += self.extract_file(filename)

        elif isinstance(node, yaml.MappingNode):
            result = {}
            for knum, val in self.construct_mapping(node).items():
                result[knum] = self.extract_file(val)

        else:
            print("Error:: unrecognised node type in !include statement")
            raise yaml.constructor.ConstructorError

        return result

    def extract_file(self, filename: str) -> dict:
        """Extract file method"""

        filepath = os.path.join(self._root, filename)
        with open(filepath, "r", encoding="utf-8") as yfile:
            return yaml.load(yfile, FmuLoader)

    # from https://gist.github.com/pypt/94d747fe5180851196eb
    def construct_mapping(self, node: _Node, deep: bool = False) -> dict:
        if not isinstance(node, MappingNode):
            raise ConstructorError(
                None,
                None,
                "Expected a mapping node, but found %s" % node.id,
                node.start_mark,
            )

        self.flatten_mapping(node)
        mapping = {}
        for key_node, value_node in node.value:
            key = self.construct_object(key_node, deep=deep)
            try:
                hash(key)
            except TypeError as exc:
                raise ConstructorError(
                    "While constructing a mapping",
                    node.start_mark,
                    "found unacceptable key (%s)" % exc,
                    key_node.start_mark,
                ) from exc
            # check for duplicate keys
            if key in mapping:
                raise ConstructorError(
                    "Found duplicate key <{}> ... {}".format(key, key_node.start_mark)
                )
            value = self.construct_object(value_node, deep=deep)
            mapping[key] = value
        return mapping



"""Testing the classes/functions in in the etc module."""

import os

import pytest
import yaml
from yaml.constructor import ConstructorError

from fmu.config import etc, utilities as util

fmux = etc.Interaction()
logger = fmux.basiclogger(__name__)

logger.info("Running tests...")


DROGON = "tests/data/yml/drogon/input/global_master_config.yml"

# always this statement
if not fmux.testsetup():
    raise SystemExit


def test_info_logger_plain():
    """Test basic logger behaviour plain, will capture output to stdin"""
    logger.info("This is a test")
    # no assert is intended


@pytest.fixture(name="mylogger")
def fixture_mylogger():
    """Add logger"""
    # need to do it like this...
    return fmux.basiclogger(__name__, level="DEBUG")


def test_info_logger(mylogger, caplog):
    """Test basic logger behaviour, will capture output to stdin"""

    mylogger.info("This is a test")
    assert "This is a test" in caplog.text

    logger.warning("This is a warning")
    assert "This is a warning" in caplog.text


def test_more_logging_tests(caplog):
    """Testing on the logging levels, see that ENV variable will override
    the basiclogger setting.
    """

    os.environ["FMU_LOGGING_LEVEL"] = "INFO"

    fmumore = etc.Interaction()  # another instance
    locallogger = fmumore.basiclogger(__name__, level="WARNING")
    locallogger.debug("Display debug")
    assert caplog.text == ""  # shall be empty
    locallogger.info("Display info")
    assert "info" in caplog.text  # INFO shall be shown, overrided by ENV!
    locallogger.warning("Display warning")
    assert "warning" in caplog.text
    locallogger.critical("Display critical")
    assert "critical" in caplog.text


def test_timer(capsys):
    """Test the timer function"""

    time1 = fmux.timer()
    for inum in range(100000):
        inum += 1

    fmux.echo(f"Used time was {fmux.timer(time1)}")
    captured = capsys.readouterr()
    assert "Used time was" in captured[0]
    # repeat to see on screen
    fmux.echo("")
    fmux.warn(f"Used time was {fmux.timer(time1)}")


def test_print_fmu_header():
    """Test writing an app header."""
    fmux.print_fmu_header("MYAPP", "0.99", info="Beta release (be careful)")


def test_user_msg():
    """Testing user messages"""

    fmux.echo("")
    fmux.echo("This is a message")
    fmux.warn("This is a warning")
    fmux.warning("This is also a warning")
    fmux.error("This is an error")
    fmux.critical("This is a critical error", sysexit=False)


def test_load_input_extended_yaml():
    """Test loading YAML will extended "fmu" option."""
    cfg = util.yaml_load(DROGON, loader="fmu")

    assert cfg["revision"] == "21.x.0.dev"
    assert (
        cfg["masterdata"]["smda"]["coordinate_system"]["uuid"]
        == "ad214d85-dac7-19da-e053-c918a4889309"
    )


def test_load_input_extended_yaml_shallfail():
    """Test loading non-standard YAML which shall fail when allow_extended is False."""
    with pytest.raises(ConstructorError, match=r"!include"):
        _ = util.yaml_load(DROGON, loader="standard")


def test_load_yaml_compare(tmp_path):
    """Test loading YAML and compare results with/without allow_extended."""

    cfg = util.yaml_load(DROGON, loader="fmu")

    yfile = tmp_path / "drogon.yml"
    with open(yfile, "w", encoding="utf-8") as stream:
        yaml.dump(cfg, stream, allow_unicode=True)

    # read a standard YAML file both with normal and extended option
    cfg1 = util.yaml_load(yfile, loader="fmu")
    cfg2 = util.yaml_load(yfile, loader="standard")

    assert cfg1 == cfg2



"""Module with some simple functions, e.g. for parsing for YAML into RMS"""

from __future__ import annotations

import yaml
from yaml.loader import Loader

from fmu.config._loader import ConstructorError, FmuLoader


def yaml_load(
    filename: str, safe: bool = True, tool: str | None = None, loader: str = "standard"
) -> dict | None:
    """Load as YAML file, return a dictionary which is the config.

    Returning an ordered dictionary is a main feature of this loader. It makes it much
    easier to compare the dictionaries returned. In addition, it allows for reading the
    input (extended) YAML format, if key ``allow_extended`` is True.

    Args:
        filename (str): Name of file (YAML formatted)
        safe (bool): If True (default), then use `safe_load` when allow_extended is
            set to False. Not applied if loader is "fmu".
        tool (str): Refers to a particular main section in the config.
            Default is None, which means 'all'.
        loader (str): If "fmu", the in-house FMU extended YAML loader that allows
            use of e.g. `!include` is applied; otherwise the default is "standard" YAML.

    Example::
        >>> import fmu.config.utilities as utils
        >>> cfg = utils.yaml_load('somefile.yml')

    """

    useloader = FmuLoader if loader.lower() == "fmu" else Loader

    with open(filename, "r", encoding="utf-8") as stream:
        try:
            if safe and loader.lower() != "fmu":
                cfg = yaml.safe_load(stream)
            else:
                cfg = yaml.load(stream, Loader=useloader)
        except ConstructorError as cerr:
            if "!include" in str(cerr):
                print(
                    "\n*** Consider setting loader='fmu' to read fmu.config "
                    "input style ***\n"
                )
            raise

    if tool is not None:
        try:
            newcfg = cfg[tool]
            cfg = newcfg
        except Exception as exc:  # pylint: disable=broad-except
            print(f"Cannot import: {exc}")
            return None

    return cfg


def compare_yaml_files(file1: str, file2: str) -> bool:
    """Compare two YAML files and return True if they are equal

    Args:
        file1 (str): Path to file1
        file2 (str): Path to file2
    """

    cfg1 = yaml_load(file1)
    cfg2 = yaml_load(file2)

    cfg1txt = yaml.dump(cfg1)
    cfg2txt = yaml.dump(cfg2)

    return cfg1txt == cfg2txt


def compare_text_files(file1: str, file2: str, comments: str = "//") -> bool:
    """Compare two text files, e.g. IPL and return True if they are equal

    Lines starting with comments indicator will be discarded

    Args:
        file1 (str): Path to file1
        file2 (str): Path to file2
        comments (str): How comment lines are indicated, e.g. "//" for IPL
    """

    text1 = ""
    text2 = ""

    with open(file1, "r", encoding="utf-8") as fil1:
        for line in fil1:
            if not line.startswith(comments):
                text1 += line

    with open(file2, "r", encoding="utf-8") as fil2:
        for line in fil2:
            if not line.startswith(comments):
                text2 += line

    return text1 == text2



"""Top-level package for fmu_config"""

try:
    from .version import __version__
except ImportError:
    __version__ = "0.0.0"

from .configparserfmu import ConfigParserFMU

__all__ = ["__version__", "ConfigParserFMU"]



"""Testing fmu-config, here focus on individual private functions"""

import datetime

import fmu.config as config
from fmu.config._configparserfmu_ipl import (
    _cast_value,
    _freeform_handle_entry,
    _guess_dtype,
)

FMUX = config.etc.Interaction()
logger = FMUX.basiclogger(__name__)

# always this statement
if not FMUX.testsetup():
    raise SystemExit


def test_ipl_guess_dtype():
    """Test IPL guess dtype output behaviour"""

    # int; single var
    var = "TESTVAR"
    entry = {var: 1}
    usekey = _guess_dtype(var, entry)
    assert isinstance(usekey, dict)
    ndict = usekey[var]
    assert isinstance(ndict, dict)
    assert ndict["dtype"] == "int"
    assert ndict["value"] == 1

    # float; single var
    var = "TESTVAR"
    entry = {var: 1.0}
    usekey = _guess_dtype(var, entry)
    assert isinstance(usekey, dict)
    ndict = usekey[var]
    assert isinstance(ndict, dict)
    assert ndict["dtype"] == "float"
    assert ndict["value"] == 1.0

    # string; single var
    var = "TESTVAR"
    entry = {var: "someword"}
    usekey = _guess_dtype(var, entry)
    assert isinstance(usekey, dict)
    ndict = usekey[var]
    assert isinstance(ndict, dict)
    assert ndict["dtype"] == "str"
    assert ndict["value"] == "someword"

    # date; single var
    var = "TESTVAR"
    entry = {var: datetime.date(1999, 11, 1)}
    usekey = _guess_dtype(var, entry)
    assert isinstance(usekey, dict)
    ndict = usekey[var]
    assert isinstance(ndict, dict)
    assert ndict["dtype"] == "date"
    assert ndict["value"] == datetime.date(1999, 11, 1)

    # int; list of
    var = "TESTVAR"
    entry = {var: [2, 1, 3, 2]}
    usekey = _guess_dtype(var, entry)
    assert isinstance(usekey, dict)
    ndict = usekey[var]
    assert isinstance(ndict, dict)
    assert ndict["dtype"] == "int"
    assert ndict["values"] == [2, 1, 3, 2]


def test_ipl_cast_value():
    """Test IPL guess dtype output behaviour"""

    value = _cast_value("233")
    assert isinstance(value, int)

    value = _cast_value("233.0")
    assert isinstance(value, float)


def test_free_form_handle_entry():
    decl, expr = _freeform_handle_entry("MYVAR", 2, None, "int", False)
    assert decl == "Int MYVAR\n"
    assert expr == "MYVAR = 2\n"

    decl, expr = _freeform_handle_entry("MYVAR", 2.0, None, "float", False)
    assert decl == "Float MYVAR\n"
    assert expr == "MYVAR = 2.0\n"



"""Testing fmu-config tools."""

import pytest

import fmu.config as config
from fmu.config import utilities as utils

# import fmu.config.fmuconfigrunner as fmurun

fmux = config.etc.Interaction()
logger = fmux.basiclogger(__name__)

REEK = "tests/data/yml/reek/global_variables.yml"

# always this statement
if not fmux.testsetup():
    raise SystemExit


def test_basic_tools():
    """Test basic tools behaviour"""

    cfg = utils.yaml_load(REEK)

    assert cfg["global"]["name"] == "Reek"

    with pytest.raises(FileNotFoundError):
        utils.yaml_load("not_a_file.xyz")



"""Script for converting the global config to various flavours of suiteble
flavours."""

from __future__ import annotations

import argparse
import os.path
import sys

import fmu.config as fmu_config
from fmu.config import etc

xfmu = etc.Interaction()
logger = xfmu.basiclogger(__name__)


def _do_parse_args(args: list[str] | None) -> argparse.Namespace:
    if args is None:
        args = sys.argv[1:]

    usetxt = "fmuconfig ... "

    parser = argparse.ArgumentParser(
        description="Configure from FMU global master", usage=usetxt
    )

    # positional:
    parser.add_argument(
        "config",
        type=str,
        help=("Input global config master file name on YAML format"),
    )

    parser.add_argument(
        "--mode",
        dest="mode",
        default="ipl",
        type=str,
        help="Mode for conversion: ipl/yaml/json/table etc...",
    )

    parser.add_argument(
        "--rootname",
        dest="rootname",
        default="global_variables",
        type=str,
        help="Root of file name",
    )

    parser.add_argument(
        "--destination",
        dest="destination",
        type=str,
        help="Destination folder (for actual values)",
    )

    parser.add_argument(
        "--template",
        dest="template",
        type=str,
        help="Template folder (for files with <xxxx> values)",
    )

    parser.add_argument(
        "--tool",
        "--entry",
        dest="tool",
        type=str,
        help="Tool or entry section to apply, e.g. rms "
        'or eclipse, or global.FWL where "." separates levels',
    )
    parser.add_argument(
        "--sep",
        dest="sep",
        type=str,
        default=",",
        help="Separator string for table mode",
    )

    if len(args) < 2:
        parser.print_help()
        print("QUIT")
        raise SystemExit

    return parser.parse_args(args)


def main(args: list[str] | None = None) -> None:
    """The fmuconfigrunner is a script that takes ..."""

    parsed_args = _do_parse_args(args)

    cfg = fmu_config.ConfigParserFMU()

    logger.info("OK %s", cfg)

    if isinstance(parsed_args.config, str):
        if not os.path.isfile(parsed_args.config):
            raise IOError("Input file does not exist")
        cfg.parse(parsed_args.config)

    if parsed_args.mode == "ipl":
        logger.info("Mode is IPL")
        cfg.to_ipl(
            rootname=parsed_args.rootname,
            destination=parsed_args.destination,
            template=parsed_args.template,
            tool=parsed_args.tool,
        )

    elif parsed_args.mode in ("yaml", "yml"):
        logger.info("Mode is YAML")
        cfg.to_yaml(
            rootname=parsed_args.rootname,
            destination=parsed_args.destination,
            template=parsed_args.template,
            tool=parsed_args.tool,
        )

    elif parsed_args.mode in ("json", "jason"):
        logger.info("Mode is JASON")
        cfg.to_json(
            rootname=parsed_args.rootname,
            destination=parsed_args.destination,
            template=parsed_args.template,
            tool=parsed_args.tool,
        )

    elif parsed_args.mode == "table":
        logger.info("Mode is TABLE")
        cfg.to_table(
            rootname=parsed_args.rootname,
            destination=parsed_args.destination,
            template=parsed_args.template,
            entry=parsed_args.tool,
            sep=parsed_args.sep,
        )
    else:
        raise RuntimeError("Invalid options for mode")


if __name__ == "__main__":
    main()



"""Testing config parsing of very small principal YAML snippets, e.g. for debugging"""

import io
from os.path import join

import pytest

import fmu.config as fcfg
from fmu.config import utilities as ut

fmux = fcfg.etc.Interaction()
logger = fmux.basiclogger(__name__)

# always this statement
if not fmux.testsetup():
    raise SystemExit


def test_dict_with_scalar_and_lists1(tmp_path):
    """s01, testing small inline snippets"""

    inp = b"""
    global:
        ADICT_WITH_SCALAR_OR_LISTS:
            1: ascalar
            3: [item1, item2, 33, 44.0]
    """

    target = tmp_path / "generic.yml"

    with target.open("w", encoding="utf-8") as out:
        out.write(inp.decode("utf-8"))

    cfg = fcfg.ConfigParserFMU()
    cfg.parse(str(target.resolve()))

    assert cfg._config["global"]["ADICT_WITH_SCALAR_OR_LISTS"][1] == "ascalar"
    assert cfg._config["global"]["ADICT_WITH_SCALAR_OR_LISTS"][3][2] == 33


def test_dict_with_scalar_and_lists2(tmp_path):
    """s02"""

    inp = b"""
    global:
        ACOMPLICATED:
            1: ascalar
            2: number ~ <>
            3: [2, 33, [12, 13, 14], manana]
    """

    target = tmp_path / "generic.yml"

    with target.open("w", encoding="utf-8") as out:
        out.write(inp.decode("utf-8"))

    cfg = fcfg.ConfigParserFMU()
    cfg.parse(str(target.resolve()))

    assert cfg._config["global"]["ACOMPLICATED"][1] == "ascalar"
    assert cfg._config["global"]["ACOMPLICATED"][3][0] == 2
    assert cfg._config["global"]["ACOMPLICATED"][3][3] == "manana"


def test_process_value1(tmp_path):
    """s03"""

    inp = b"""
    global:
        WELLCONCEPT: REV1_A15 ~ <WELL_CONS>
    """

    target = tmp_path / "generic.yml"
    with target.open("w", encoding="utf-8") as out:
        out.write(inp.decode("utf-8"))

    cfg = fcfg.ConfigParserFMU()
    cfg.parse(str(target.resolve()))

    outfolder = tmp_path
    out = str(outfolder.resolve())

    cfg.to_yaml(rootname="s03", destination=out, template=out)

    actual = ut.yaml_load(join(out, "s03.yml"))

    assert actual["global"]["WELLCONCEPT"] == "REV1_A15"

    tmpl = ut.yaml_load(join(out, "s03.yml.tmpl"))

    assert tmpl["global"]["WELLCONCEPT"] == "<WELL_CONS>"


def test_process_value2(tmp_path):
    """s04"""

    inp = b"""
    global:
        WHATEVER: &onx ONX ~ <ONX>
        WHATEVER2: *onx
        NUMBER1: 2.0 ~ <NUM1>
    """

    target = tmp_path / "generic.yml"
    with target.open("w", encoding="utf-8") as out:
        out.write(inp.decode("utf-8"))

    cfg = fcfg.ConfigParserFMU()
    cfg.parse(str(target))

    out = str(tmp_path.resolve())

    cfg.to_yaml(rootname="s04", destination=out, template=out)

    actual = ut.yaml_load(join(out, "s04.yml"))

    assert actual["global"]["WHATEVER"] == "ONX"
    assert actual["global"]["NUMBER1"] == 2.0

    tmpl = ut.yaml_load(join(out, "s04.yml.tmpl"))

    assert tmpl["global"]["WHATEVER"] == "<ONX>"


@pytest.mark.parametrize("exponent", range(16))
def test_small_float(tmp_path, exponent):
    """Test for issue 1, small floats in input yaml can give non-compliant
    formats in output IPL (IPL does not support Pythons scientific format)

    IPL accepts 1.0E-4, 1.0e-4, 1.0e-04, but NOT 1E-3 or 1e-4, i.e. there must be
    a dot after '1'"""

    inp = (
        b"""
    rms:
        SMALLFLOAT: 1.0E-%d"""
        % exponent
    )

    target = tmp_path / "generic.yml"
    with io.open(str(target.resolve()), "wb") as out:
        out.write(inp)

    cfg = fcfg.ConfigParserFMU()
    cfg.parse(str(target))

    out = str(tmp_path.resolve())

    cfg.to_ipl(rootname="foo", destination=out, template=out)
    with open(join(str(tmp_path.resolve()), "foo.ipl"), encoding="utf-8") as f:
        ipl_lines = [line for line in f.readlines() if line.startswith("SMALLFLOAT")]
    significand = ipl_lines[0].split(" = ")[1].split("E")[0].split("e")[0]

    # Verify we have a floating point significand if we have written
    # exponential notation
    if "e" in ipl_lines[0].lower():
        assert len(significand) > 1
        assert "." in significand


def test_filepath(tmp_path):
    """Test when filepath is present as substitute, s05"""

    inp = b"""
    global:
        MYFILE: /tmp/fil ~ <FILE>
    """

    target = tmp_path / "generic.yml"
    with target.open("w", encoding="utf-8") as out:
        out.write(inp.decode("utf-8"))

    cfg = fcfg.ConfigParserFMU()
    cfg.parse(str(target))

    out = str(tmp_path.resolve())

    cfg.to_yaml(rootname="s05", destination=out, template=out)

    actual = ut.yaml_load(join(out, "s05.yml"))

    assert actual["global"]["MYFILE"] == "/tmp/fil"

    tmpl = ut.yaml_load(join(out, "s05.yml.tmpl"))

    assert tmpl["global"]["MYFILE"] == "<FILE>"
